from pathlib import Path
from tempfile import NamedTemporaryFile

from minio_manager.classes.errors import MinioInvalidIamCredentialsError
from minio_manager.classes.mc_wrapper import McWrapper
from minio_manager.classes.minio_resources import ServiceAccount
from minio_manager.classes.secrets import MinioCredentials
from minio_manager.clients import get_mc_wrapper, get_minio_config, get_secret_manager
from minio_manager.utilities import logger


def service_account_exists(client: McWrapper, credentials: MinioCredentials):
    try:
        if credentials.access_key:
            client.service_account_info(credentials.access_key)
            return True
    except MinioInvalidIamCredentialsError as e:
        # account does not exist in MinIO
        logger.debug(f"Error for {credentials.access_key}: {e}")

    logger.debug(f"Access key for {credentials.name} not found in secret backend, trying to find it in MinIO.")
    sa_list = client.service_account_list(client.cluster_controller_user)
    for sa in sa_list:
        sa_info = client.service_account_info(sa.accessKey)
        if hasattr(sa_info, "name") and sa_info.name == credentials.name:
            logger.debug(f"Found access key '{sa_info.accessKey}' for '{credentials.name}' in MinIO: {sa_info}")
            return True

    return False


def generate_service_account_policy(account: ServiceAccount) -> Path:
    with Path("examples/service-account-policy-base.json").open() as base:
        base_policy = base.read()

    temp_file = NamedTemporaryFile(prefix=account.bucket, suffix=".json", delete=False)
    with temp_file as out:
        new_content = base_policy.replace("BUCKET_NAME_REPLACE_ME", account.bucket)
        out.write(new_content.encode("utf-8"))

    return Path(temp_file.name)


def handle_service_account(account: ServiceAccount):
    """
    Manage service accounts.

    Steps taken:
    1) check if service account exists in minio
    2) check if service account exists in secret backend
    3) if it exists in MinIO but not the secret backend, throw an error and return
    4) if it exists in the secret backend but not in MinIO, create it using the secret backend credentials
    5) if it does not, create service account in minio and secret backend

    Args:
        account (ServiceAccount)
    """
    client = get_mc_wrapper()
    secrets = get_secret_manager(get_minio_config())

    # Determine if access key credentials exists in secret backend
    credentials = secrets.get_credentials(account.name)
    # Determine if access key exists in MinIO
    sa_exists = service_account_exists(client, credentials)

    # Scenario 1: service account exists in MinIO but not in secret backend
    if sa_exists and not credentials.access_key:
        logger.error(
            f"Service account {account.name} exists in MinIO but not in secret backend! Manual intervention required."
        )
        logger.error(
            "Either find the credentials elsewhere and add them to the secret backend, or delete the service "
            "account from MinIO and try again."
        )
        return

    # Scenario 2: service account exists in secret backend but not in MinIO
    if credentials.secret_key and not sa_exists:
        logger.warning(
            f"Service account {account.name} exists in secret backend but not in MinIO. Using existing credentials."
        )
        client.service_account_add(credentials)
        sa_exists = True
        logger.info(f"Created service account '{credentials.name}', access key: {credentials.access_key}")

    # Scenario 3: service account does not exist in neither MinIO nor the secret backend
    if not sa_exists and not credentials.access_key:
        # TODO: catch scenario where an access key is deleted in MinIO, but MinIO does not accept the creation of a
        #  service account with the same access key.
        # Create the service account in MinIO
        credentials = client.service_account_add(credentials)
        # Create credentials in the secret backend
        secrets.set_password(credentials)
        logger.info(f"Created service account '{credentials.name}' with access key '{credentials.access_key}'")

    if account.bucket:
        # TODO: validate existing policy before assigning
        policy_file = generate_service_account_policy(account)
        client.service_account_set_policy(credentials.access_key, str(policy_file))
        policy_file.unlink()



from minio_manager.classes.minio_resources import Bucket, ServiceAccount
from minio_manager.clients import get_s3_client
from minio_manager.service_account_handler import handle_service_account
from minio_manager.utilities import logger


def configure_versioning(client, bucket):
    versioning_status = client.get_bucket_versioning(bucket.name)
    if versioning_status.status != bucket.versioning.status:
        client.set_bucket_versioning(bucket.name, bucket.versioning)
        logger.debug(f"Versioning {bucket.versioning.status.lower()} for bucket {bucket.name}")


def configure_lifecycle(client, bucket):
    lifecycle_status = client.get_bucket_lifecycle(bucket.name)
    if lifecycle_status != bucket.lifecycle_config:
        client.set_bucket_lifecycle(bucket.name, bucket.lifecycle_config)
        logger.debug(f"Lifecycle {bucket.lifecycle_config} for bucket {bucket.name}")


def handle_bucket(bucket: Bucket):
    """Handle the specified bucket.

    First validates the existence of the bucket. If it does not exist, it will be created.
    Then it will compare the current bucket versioning configuration and update it if needed.
    Lastly, a related service account will be created that gives access to this specific bucket.

    Args:
        bucket (Bucket): The bucket to handle.
    """
    client = get_s3_client()
    if not client.bucket_exists(bucket.name):
        logger.info("Creating bucket %s" % bucket.name)
        client.make_bucket(bucket.name)
    else:
        logger.info(f"Bucket {bucket.name} already exists")

    configure_versioning(client, bucket)
    configure_lifecycle(client, bucket)

    if bucket.create_sa:
        # TODO: is there a nicer way to go about this?
        service_account = ServiceAccount(bucket.name, bucket.name)
        handle_service_account(service_account)



class MinioManagerBaseError(Exception):
    """Base class for Minio Manager errors."""

    def __init__(self, message: str, cause=None):
        super().__init__(f"{message}: {cause}" if cause else message)


class MinioInvalidIamCredentialsError(MinioManagerBaseError):
    """Raised when invalid IAM credentials are used."""


class MinioInvalidAccessKeyId(MinioManagerBaseError):
    """Raised when an invalid user access key ID is provided."""


class MinioInternalError(MinioManagerBaseError):
    """Raised when a MinIO internal error occurred."""


class MinioNoSuchUserError(MinioManagerBaseError):
    """Raised when a specified user account is not found."""


class MinioNoSuchPolicyError(MinioManagerBaseError):
    """Raised when a specified policy is not found."""


class MinioInvalidSecretKeyError(MinioManagerBaseError):
    """Raised when an invalid secret key is used."""


class MinioInvalidAccessKeyError(MinioManagerBaseError):
    """Raised when an invalid access key is used."""


class MinioIamServiceAccountNotAllowedError(MinioManagerBaseError):
    """Raised when trying to add a service account that already exists."""


error_map = {
    "XMinioInvalidIAMCredentials": MinioInvalidIamCredentialsError,
    "InvalidAccessKeyId": MinioInvalidAccessKeyId,
    "XMinioIAMServiceAccountNotAllowed": MinioIamServiceAccountNotAllowedError,
    "InternalError": MinioInternalError,
    "XMinioAdminNoSuchUser": MinioNoSuchUserError,
    "XMinioAdminNoSuchPolicy": MinioNoSuchPolicyError,
    "XMinioAdminInvalidSecretKey": MinioInvalidSecretKeyError,
    "XMinioAdminInvalidAccessKey": MinioInvalidAccessKeyError,
}


def raise_specific_error(error_code, error_message):
    if error_code not in error_map:
        raise MinioManagerBaseError(error_code, error_message)
    raise error_map[error_code](error_message)



from __future__ import annotations

from minio.lifecycleconfig import LifecycleConfig
from minio.versioningconfig import VersioningConfig


class Bucket:
    def __init__(
        self,
        name: str,
        create_service_account: bool = True,
        versioning: VersioningConfig | None = None,
        lifecycle_config: LifecycleConfig | None = None,
    ):
        self.name = name
        self.create_sa = create_service_account
        self.versioning = versioning
        self.lifecycle_config = lifecycle_config


class BucketPolicy:
    def __init__(self, bucket: str, policy_file: str):
        self.bucket = bucket
        self.policy_file = policy_file


class ServiceAccount:
    def __init__(self, name: str, bucket: str | None = None):
        self.name = name
        self.bucket = bucket


class IamPolicy:
    def __init__(self, name: str, policy_file: str):
        self.name = name
        self.policy_file = policy_file


class IamPolicyAttachment:
    def __init__(self, username: str, policies: list):
        self.username = username
        self.policies = policies



import re
from logging import Filter, LogRecord
from re import Pattern


class MinioManagerFilter(Filter):
    wrapper_secret_re = re.compile(r"--secret-key (?P<secret>[\w+/]*)")
    alias_set_secret_re = re.compile(r"alias set .+ (?P<secret>[\w+/]*)$")

    def filter(self, record: LogRecord) -> bool:
        if "--secret" in record.msg:
            record.msg = self.mask_secret(record.msg, self.wrapper_secret_re)
        if "alias set" in record.msg:
            record.msg = self.mask_secret(record.msg, self.alias_set_secret_re)

        return True

    @staticmethod
    def mask_secret(message: str, regex: Pattern) -> str:
        result = regex.search(message)
        if result:
            message = message.replace(result.group("secret"), "************")
        return message



from __future__ import annotations

import json
from pathlib import Path

from minio.commonconfig import Filter
from minio.lifecycleconfig import Expiration, LifecycleConfig, NoncurrentVersionExpiration, Rule
from minio.versioningconfig import VersioningConfig

from minio_manager.classes.minio_resources import Bucket, BucketPolicy, IamPolicy, IamPolicyAttachment, ServiceAccount
from minio_manager.utilities import logger, read_yaml, retrieve_environment_variable

default_bucket_versioning = retrieve_environment_variable("MINIO_MANAGER_DEFAULT_BUCKET_VERSIONING", "Suspended")
default_bucket_lifecycle_policy = retrieve_environment_variable("MINIO_MANAGER_DEFAULT_LIFECYCLE_POLICY", None)


class ClusterResources:
    """MinIO Cluster configuration object, aka the cluster contents: buckets, policies, etc."""

    buckets: list[Bucket]
    bucket_policies: list[BucketPolicy]
    service_accounts: list[ServiceAccount]
    iam_policies: list[IamPolicy]
    iam_policy_attachments: list[IamPolicyAttachment]

    def parse_buckets(self, buckets):
        """Parse the provided buckets with the following steps:
        For each provided bucket
            1. check the provided versioning. If versioning is not provided, set the default.
            2. check if an object lifecycle JSON file is provided, use the default_bucket_lifecycle_policy, or skip OLM
            3. parse the file and create a LifecycleConfig object for the bucket
            4. create a Bucket object

        Args:
            buckets: list of buckets to parse

        Returns:
            [Bucket]: list of Bucket objects
        """
        if not buckets:
            logger.info("No buckets configured, skipping.")
            return []

        bucket_objects = []
        try:
            for bucket in buckets:
                vs = bucket.get("versioning", None)
                vc = VersioningConfig(vs) if vs else VersioningConfig(default_bucket_versioning)
                create_sa = bucket.get("create_service_account", True)
                lifecycle_file = bucket.get("object_lifecycle_file", default_bucket_lifecycle_policy)
                lifecycle_config = self.parse_bucket_lifecycle_file(lifecycle_file)
                bucket_objects.append(Bucket(bucket["name"], create_sa, vc, lifecycle_config))
        except AttributeError:
            logger.info("No buckets configured, skipping.")

        return bucket_objects

    def parse_bucket_lifecycle_file(self, lifecycle_file: str) -> LifecycleConfig | None:
        """Parse a list of bucket lifecycle config files.
        The config files must be in JSON format and can be best obtained by running the following command:
            mc ilm rule export $cluster/$bucket > $policy_file.json

        Args:
            lifecycle_file: list of lifecycle config files

        Returns:
            LifecycleConfig object
        """
        if not lifecycle_file:
            return

        rules: list = []

        with Path(lifecycle_file).open() as f:
            config_data = json.load(f)
            for rule_data in config_data["Rules"]:
                parsed_rule = self.parse_bucket_lifecycle_rule(rule_data)
                rules.append(parsed_rule)

        if not rules:
            return

        return LifecycleConfig(rules)

    @staticmethod
    def parse_bucket_lifecycle_rule(rule_data) -> Rule:
        """Parse a single bucket object lifecycle rule
        TODO: implement date and days in Expiration, implement Transition, NoncurrentVersionTransition, Filter, and
          AbortIncompleteMultipartUpload

        Args:
            rule_data: dict with rule data

        Returns:
            Rule
        """
        rule_dict = {"status": rule_data.get("Status"), "rule_id": rule_data.get("ID")}

        expiration = rule_data.get("Expiration", None)
        if expiration:
            expire_delete_marker = expiration.get("ExpiredObjectDeleteMarker", None)
            rule_dict["expiration"] = Expiration(expired_object_delete_marker=expire_delete_marker)

        noncurrent_version_expiration = rule_data.get("NoncurrentVersionExpiration", None)
        if noncurrent_version_expiration:
            noncurrent_expire_days = noncurrent_version_expiration.get("NoncurrentDays")
            rule_dict["noncurrent_version_expiration"] = NoncurrentVersionExpiration(noncurrent_expire_days)

        # An empty filter is required for the rule to be valid
        rule_dict["rule_filter"] = Filter(prefix="")

        rule = Rule(**rule_dict)
        return rule

    @staticmethod
    def parse_bucket_policies(bucket_policies):
        if not bucket_policies:
            logger.info("No bucket policies configured, skipping.")
            return []

        bucket_policy_objects = []
        for bucket_policy in bucket_policies:
            bucket_policy_objects.append(BucketPolicy(bucket_policy["bucket"], bucket_policy["policy_file"]))

        return bucket_policy_objects

    @staticmethod
    def parse_service_accounts(service_accounts):
        if not service_accounts:
            logger.info("No service accounts configured, skipping.")
            return []

        service_account_objects = []
        for service_account in service_accounts:
            bucket = service_account.get("bucket", None)
            service_account_objects.append(ServiceAccount(service_account["name"], bucket))

        return service_account_objects

    @staticmethod
    def parse_iam_policy_attachments(iam_policy_attachments):
        if not iam_policy_attachments:
            logger.info("No IAM policy attachments configured, skipping.")
            return []

        iam_policy_attachment_objects = []
        for user in iam_policy_attachments:
            iam_policy_attachments.append(IamPolicyAttachment(user["username"], user["policies"]))

        return iam_policy_attachment_objects

    @staticmethod
    def parse_iam_policies(iam_policies):
        if not iam_policies:
            logger.info("No IAM policies configured, skipping.")
            return []

        iam_policy_objects = []
        for iam_policy in iam_policies:
            iam_policy_objects.append(IamPolicy(iam_policy["name"], iam_policy["policy_file"]))

        return iam_policy_objects

    def parse_resources(self, resources_file: dict):
        resources_file = read_yaml(resources_file)

        buckets = resources_file.get("buckets", [])
        self.buckets = self.parse_buckets(buckets)

        bucket_policies = resources_file.get("bucket_policies", [])
        self.bucket_policies = self.parse_bucket_policies(bucket_policies)

        service_accounts = resources_file.get("service_accounts", [])
        self.service_accounts = self.parse_service_accounts(service_accounts)

        iam_policies = resources_file.get("iam_policies", [])
        self.iam_policies = self.parse_iam_policies(iam_policies)

        iam_policy_attachments = resources_file.get("iam_policy_attachments", [])
        self.iam_policy_attachments = self.parse_iam_policy_attachments(iam_policy_attachments)


class MinioConfig:
    """MinIO server configuration object, the connection details."""

    def __init__(self):
        self.name = retrieve_environment_variable("MINIO_MANAGER_CLUSTER_NAME")
        self.endpoint = retrieve_environment_variable("MINIO_MANAGER_S3_ENDPOINT")
        self.secure = retrieve_environment_variable("MINIO_MANAGER_S3_ENDPOINT_SECURE", True)
        self.controller_user = retrieve_environment_variable("MINIO_MANAGER_MINIO_CONTROLLER_USER")
        self.access_key = None
        self.secret_key = None
        self.cluster_resources = retrieve_environment_variable("MINIO_MANAGER_CLUSTER_RESOURCES_FILE")
        self.secret_backend_type = retrieve_environment_variable("MINIO_MANAGER_SECRET_BACKEND_TYPE")
        self.secret_s3_bucket = retrieve_environment_variable("MINIO_MANAGER_SECRET_BACKEND_S3_BUCKET")



from __future__ import annotations

from minio import Minio, MinioAdmin, credentials

from minio_manager.classes.config import MinioConfig
from minio_manager.classes.mc_wrapper import McWrapper
from minio_manager.classes.secrets import SecretManager
from minio_manager.utilities import logger

config, s3_client, admin_client, mc_wrapper, secrets = None, None, None, None, None


def get_s3_client() -> Minio:
    """Set up MinIO S3 client for the specified cluster.

    Returns:
        Minio: MinIO S3 client

    """
    global s3_client
    if s3_client:
        return s3_client
    c = get_minio_config()
    s3_client = Minio(endpoint=c.endpoint, access_key=c.access_key, secret_key=c.secret_key, secure=c.secure)
    return s3_client


def get_minio_admin_client() -> MinioAdmin:
    """Set up MinIO Admin client

    Returns:
        MinioAdmin: MinIO Admin client
    """
    global admin_client
    if admin_client:
        return admin_client
    c = get_minio_config()
    provider = credentials.StaticProvider(c.access_key, c.secret_key)
    admin_client = MinioAdmin(endpoint=c.endpoint, credentials=provider, secure=c.secure)
    return admin_client


def get_mc_wrapper() -> McWrapper:
    """Set up 'mc' wrapper

    Returns:
        McWrapper: MinIO CLI client wrapper
    """
    global mc_wrapper
    if mc_wrapper:
        return mc_wrapper
    c = get_minio_config()
    mc_wrapper = McWrapper(config=c)
    return mc_wrapper


def get_secret_manager(c: MinioConfig) -> SecretManager:
    global secrets
    if not secrets:
        secrets = SecretManager(c)
    return secrets


def get_minio_config():
    global config
    if not config:
        logger.info("Configuring...")
        config = MinioConfig()
    return config



import json

from minio import S3Error
from minio.error import MinioAdminException

from minio_manager.classes.minio_resources import BucketPolicy, IamPolicy, IamPolicyAttachment
from minio_manager.clients import get_minio_admin_client, get_s3_client
from minio_manager.utilities import logger, read_json, sort_policy


def handle_bucket_policy(bucket_policy: BucketPolicy):
    """
    Manage policies for buckets.

    If the policy doesn't exist, create it.
    If the policy exists, compare the desired policy with the current policy, and update if needed.

    Args:
        bucket_policy: BucketPolicy
    """
    client = get_s3_client()
    current_policy = None
    desired_policy = sort_policy(read_json(bucket_policy.policy_file))
    desired_policy_json = json.dumps(desired_policy)

    try:
        current_policy_str = client.get_bucket_policy(bucket_policy.bucket)
        current_policy = sort_policy(json.loads(current_policy_str))
        logger.debug(f"Current policy: {current_policy}")
        logger.debug(f"Desired policy: {desired_policy}")
    except S3Error as s3e:
        if s3e.code == "NoSuchBucketPolicy":
            logger.info(f"Creating bucket policy for {bucket_policy.bucket}")
            try:
                client.set_bucket_policy(bucket_policy.bucket, desired_policy_json)
                current_policy = client.get_bucket_policy(bucket_policy.bucket)
            except S3Error as sbe:
                if sbe.code == "MalformedPolicy":
                    logger.exception("Do the resources in the policy file match the bucket name? Is it valid JSON?")
                    return

    if current_policy == desired_policy:
        return

    logger.info(f"Desired bucket policy for '{bucket_policy.bucket}' does not match current policy. Updating.")
    try:
        client.set_bucket_policy(bucket_policy.bucket, desired_policy_json)
    except S3Error:
        logger.exception("Failed to update bucket policy")


def handle_iam_policy(iam_policy: IamPolicy):
    """
    Manage IAM policies for users.
    If the policy doesn't exist, create it.
    If the policy exists, compare the desired policy with the current policy, and update if needed.

    Args:
        iam_policy: IamPolicy
    """
    client = get_minio_admin_client()
    current_policy = None
    desired_policy = sort_policy(read_json(iam_policy.policy_file))

    try:
        current_policy_str = client.policy_info(iam_policy.name)
        current_policy = sort_policy(json.loads(current_policy_str))
        logger.debug(f"Current (sorted) policy: {current_policy}")
        logger.debug(f"Desired (sorted) policy: {desired_policy}")
    except MinioAdminException as mae:
        # noinspection PyProtectedMember
        mae_obj = json.loads(mae._body)
        if mae_obj["Code"] == "XMinioAdminNoSuchPolicy":
            logger.info(f"IAM policy {iam_policy.name} does not exist, creating.")
            client.policy_add(iam_policy.name, iam_policy.policy_file)
            current_policy = client.policy_info(iam_policy.name)
        else:
            logger.exception("An unknown exception occurred")

    if current_policy == desired_policy:
        return

    logger.info(f"Desired IAM policy '{iam_policy.name}' does not match current policy. Updating IAM policy.")
    client.policy_add(iam_policy.name, iam_policy.policy_file)


def handle_iam_policy_attachments(user: IamPolicyAttachment):
    """
    Manage user policy attachments.

    Args:
        user: IamPolicyAttachment
    """
    client = get_minio_admin_client()
    logger.debug(f"Handling user policy attachments for '{user.username}'")
    for policy_name in user.policies:
        logger.debug(f"Attaching policy '{policy_name}' to access key '{user.username}'")
        client.policy_set(policy_name, user.username)

    # TODO: don't set the attachments if they're already attached



from minio_manager.bucket_handler import handle_bucket
from minio_manager.classes.config import ClusterResources
from minio_manager.policy_handler import handle_bucket_policy, handle_iam_policy, handle_iam_policy_attachments
from minio_manager.service_account_handler import handle_service_account
from minio_manager.utilities import logger


def handle_resources(resources: ClusterResources):
    """Handle the provided bucket, bucket policies, IAM policies, and user policy attachments, in that order.

    Args:
        resources: ClusterResources object with all resources
    """
    logger.info("Handling buckets...")
    for bucket in resources.buckets:
        handle_bucket(bucket)

    logger.info("Handling bucket policies...")
    for bucket_policy in resources.bucket_policies:
        handle_bucket_policy(bucket_policy)

    logger.info("Handling service accounts...")
    for service_account in resources.service_accounts:
        handle_service_account(service_account)

    logger.info("Handling IAM policies...")
    for iam_policy in resources.iam_policies:
        handle_iam_policy(iam_policy)

    logger.info("Handling IAM policy attachments...")
    for iam_policy_attachment in resources.iam_policy_attachments:
        handle_iam_policy_attachments(iam_policy_attachment)



import json
import os
import shutil
import subprocess
from pathlib import Path
from types import SimpleNamespace

from minio_manager.classes.config import MinioConfig
from minio_manager.classes.errors import MinioManagerBaseError, raise_specific_error
from minio_manager.classes.secrets import MinioCredentials
from minio_manager.utilities import logger


class McWrapper:
    def __init__(self, config: MinioConfig, timeout=60):
        logger.debug("Initialising McWrapper")
        self.cluster_name = config.name
        self.cluster_controller_user = config.controller_user
        self.timeout = timeout
        self.mc_config_path = self.set_config_path()
        self.mc = self.find_mc_command()
        self.configure(config.endpoint, config.access_key, config.secret_key, config.secure)
        logger.info("Initialised McWrapper")

    def _run(self, args, multiline=False):
        """Execute mc command and return JSON output."""
        logger.debug(f"Running: {self.mc} --json {' '.join(args)}")
        proc = subprocess.run(
            [self.mc, "--json", *args],  # noqa: S603
            capture_output=True,
            timeout=self.timeout,
            text=True,
        )
        if not proc.stdout:
            return [] if multiline else {}
        if multiline:
            return [json.loads(line, object_hook=lambda d: SimpleNamespace(**d)) for line in proc.stdout.splitlines()]
        return json.loads(proc.stdout, object_hook=lambda d: SimpleNamespace(**d))

    @staticmethod
    def set_config_path():
        """Set the path to the mc config.json file"""
        env_mc_config_path = os.getenv("MC_CONFIG_PATH")
        env_home = os.getenv("HOME")
        mc_paths = [
            f"{env_mc_config_path}/config.json",
            f"{env_home}/.mc/config.json",
            f"{env_home}/.mcli/config.json",
        ]
        for path in mc_paths:
            if os.path.exists(path):
                return path

    @staticmethod
    def find_mc_command() -> Path:
        """Configure the path to the mc command, as it may be named 'mcli' on some systems."""
        mc = shutil.which("mc")
        if not mc:
            mc = shutil.which("mcli")
        return Path(mc)

    def configure(self, endpoint, access_key, secret_key, secure: bool):
        """Ensure the proper alias is configured for the cluster."""
        logger.debug(f"Validating config for cluster {self.cluster_name}")
        cluster_ready = self._run(["ready", self.cluster_name])
        logger.debug(f"Cluster status: {cluster_ready}")
        if not cluster_ready.error:
            # Cluster is configured & available
            return

        logger.info("Endpoint is not configured or erroneous, configuring...")
        url = f"https://{endpoint}" if secure else f"http://{endpoint}"
        alias_set_resp = self._run(["alias", "set", self.cluster_name, url, access_key, secret_key])
        if hasattr(alias_set_resp, "error"):
            error_details = alias_set_resp.error.cause.error
            try:
                raise_specific_error(error_details.Code, error_details.Message)
            except AttributeError as ae:
                logger.exception("Unknown error!")
                raise MinioManagerBaseError(alias_set_resp.error.cause.message) from ae

        cluster_ready = self._run(["ready", self.cluster_name])
        if cluster_ready.healthy:
            # Cluster is configured & available
            return

        if hasattr(cluster_ready, "error"):
            # A connection error occurred
            raise ConnectionError(cluster_ready.error)

    def _service_account_run(self, cmd, args):
        """
        mc admin user svcacct helper function, no need to specify the cluster name
        Args:
            cmd: str, the svcacct command
            args: list of arguments to the command

        Returns: a SimpleNamespace object

        """
        multiline = cmd in ["list", "ls"]
        resp = self._run(["admin", "user", "svcacct", cmd, self.cluster_name, *args], multiline=multiline)
        if hasattr(resp, "error") and resp.error:
            error_details = resp.error.cause.error
            raise_specific_error(error_details.Code, error_details.Message)
        return resp

    def service_account_add(self, credentials: MinioCredentials) -> MinioCredentials:
        """
        mc admin user svcacct add alias-name 'username' --name "sa-test-key"

        Args:
            credentials (MinioCredentials): object containing at least the user-friendly name of the service account

        Returns: MinioCredentials with the access and secret keys added to it
        """
        # Create the service account in MinIO
        args = [self.cluster_controller_user, "--name", credentials.name]
        if credentials.secret_key:
            args.extend(["--secret-key", credentials.secret_key])
        if credentials.access_key:
            args.extend(["--access-key", credentials.access_key])
        resp = self._service_account_run("add", args)
        credentials.access_key = resp.accessKey
        credentials.secret_key = resp.secretKey
        return credentials

    def service_account_list(self, access_key):
        """mc admin user svcacct ls alias-name 'access_key'"""
        return self._service_account_run("ls", [access_key])

    def service_account_info(self, access_key):
        """mc admin user svcacct info alias-name service-account-access-key"""
        return self._service_account_run("info", [access_key])

    def service_account_delete(self):
        """mc admin user svcacct rm alias-name service-account-access-key"""
        raise NotImplementedError

    def service_account_set_policy(self, access_key, policy_file):
        """mc admin user svcacct edit alias-name service-account-access-key --policy policy-file"""
        return self._service_account_run("edit", [access_key, "--policy", policy_file])



import time

from dotenv import load_dotenv

from minio_manager.classes.config import ClusterResources
from minio_manager.clients import get_minio_config, get_secret_manager
from minio_manager.resource_handler import handle_resources
from minio_manager.utilities import logger


def main():
    load_dotenv()
    logger.info("Starting MinIO Manager...")
    start_time = time.time()
    config = get_minio_config()

    logger.info("Loading resources...")
    cluster_resources = ClusterResources()
    cluster_resources.parse_resources(config.cluster_resources)

    logger.info("Loading secret backend...")
    s = get_secret_manager(config)
    run_user_credentials = s.get_credentials(config.controller_user)
    config.access_key = run_user_credentials.access_key
    config.secret_key = run_user_credentials.secret_key

    logger.info("Handling cluster resources...")
    handle_resources(cluster_resources)

    secrets = get_secret_manager(config)
    secrets.cleanup()
    end_time = time.time()
    logger.info(f"Execution took {end_time - start_time:.2f} seconds.")



import json
import logging
import os

import yaml

from minio_manager.classes.logging_config import MinioManagerFilter

logger = None


def read_yaml(file):
    with open(file) as f:
        return yaml.safe_load(f)


def read_json(file) -> dict:
    with open(file) as f:
        return json.load(f)


def setup_logging():
    global logger
    log_level = retrieve_environment_variable("MINIO_MANAGER_LOG_LEVEL", "INFO")
    logger = logging.getLogger("minio-manager") if log_level != "DEBUG" else logging.getLogger("root")
    handler = logging.StreamHandler()
    minio_manager_filter = MinioManagerFilter()

    if log_level == "DEBUG":
        logger.setLevel(logging.DEBUG)
        log_format = "[{asctime}] [{levelname:^8s}] [{filename:>26s}:{lineno:<4d} - {funcName:<24s} ] {message}"
    else:
        logger.setLevel(logging.INFO)
        log_format = "[{asctime}] [{levelname:^8s}] {message}"

    formatter = logging.Formatter(log_format, style="{")
    handler.setFormatter(formatter)
    handler.addFilter(minio_manager_filter)
    logger.addHandler(handler)
    logger.debug(f"Configured log level: {log_level}")


def sort_policy(policy: dict):
    """

    Args:
        policy: dict
            MinIO policy documents use the same schema as AWS IAM Policy documents.
            https://docs.aws.amazon.com/IAM/latest/UserGuide/access.html

    Returns: dict with sorted Actions and Resources
    """
    for index, statement in enumerate(policy["Statement"]):
        statement["Action"] = sorted(statement["Action"])
        statement["Resource"] = sorted(statement["Resource"])
        # TODO: also sort Principals
        statement = dict(sorted(statement.items()))
        policy["Statement"][index] = statement
    return policy


def retrieve_environment_variable(name: str, default=None) -> str:
    """
    Get an environment variable and strip any leading and trailing single and double quotes.
    This is because Python apparently literally loads them.
    Args:
        default: str, default if not set
        name: str, the name of the environment variable

    Returns: str stripped
    """
    try:
        variable = os.environ[name]
        strip_double_quotes = variable.strip('"')
        return strip_double_quotes.strip("'")
    except KeyError:
        if not default:
            logger.critical(f"Required environment variable {name} not found!")
            exit(1)

    return default


if not logger:
    setup_logging()
